@@ -10,14 +10,12 @@
 
 GlobalState global_state = {{0}};
 
-#define REMOVER(array, qid, count, structure, destroy, capacity) { \
+#define REMOVER(array, qid, count, destroy, capacity) { \
     for (size_t i = 0; i < count; i++) { \
         if (array[i].id == qid) { \
             destroy(array + i); \
-            memset(array + i, 0, sizeof(structure)); \
-            size_t num_to_right = count - 1 - i; \
-            if (num_to_right) memmove(array + i, array + i + 1, num_to_right * sizeof(structure)); \
-            (count)--; \
+            zero_at_i(array, i); \
+            remove_i_from_array(array, i, count); \
             break; \
         } \
     }}
@@ -73,15 +71,65 @@ os_window_for_kitty_window(id_type kitty_window_id) {
     return NULL;
 }
 
+static void
+send_bgimage_to_gpu(BackgroundImageLayout layout, BackgroundImage *bgimage) {
+    RepeatStrategy r;
+    switch (layout) {
+        case SCALED:
+            r = REPEAT_CLAMP; break;
+        case MIRRORED:
+            r = REPEAT_MIRROR; break;
+        case TILING:
+        default:
+            r = REPEAT_DEFAULT; break;
+    }
+    bgimage->texture_id = 0;
+    send_image_to_gpu(&bgimage->texture_id, bgimage->bitmap, bgimage->width,
+            bgimage->height, false, true, OPT(background_image_linear), r);
+    free(bgimage->bitmap); bgimage->bitmap = NULL;
+}
+
+static void
+free_bgimage(BackgroundImage **bgimage, bool release_texture) {
+    if (*bgimage && (*bgimage)->refcnt) {
+        (*bgimage)->refcnt--;
+        if ((*bgimage)->refcnt == 0) {
+            free((*bgimage)->bitmap); (*bgimage)->bitmap = NULL;
+            if (release_texture) free_texture(&(*bgimage)->texture_id);
+            free(*bgimage);
+        }
+    }
+    bgimage = NULL;
+}
+
 OSWindow*
 add_os_window() {
     WITH_OS_WINDOW_REFS
     ensure_space_for(&global_state, os_windows, OSWindow, global_state.num_os_windows + 1, capacity, 1, true);
     OSWindow *ans = global_state.os_windows + global_state.num_os_windows++;
-    memset(ans, 0, sizeof(OSWindow));
+    zero_at_ptr(ans);
     ans->id = ++global_state.os_window_id_counter;
     ans->tab_bar_render_data.vao_idx = create_cell_vao();
+    ans->gvao_idx = create_graphics_vao();
     ans->background_opacity = OPT(background_opacity);
+
+    bool wants_bg = OPT(background_image) && OPT(background_image)[0] != 0;
+    if (wants_bg) {
+        if (!global_state.bgimage) {
+            global_state.bgimage = calloc(1, sizeof(BackgroundImage));
+            if (!global_state.bgimage) fatal("Out of memory allocating the global bg image object");
+            global_state.bgimage->refcnt++;
+            size_t size;
+            if (png_path_to_bitmap(OPT(background_image), &global_state.bgimage->bitmap, &global_state.bgimage->width, &global_state.bgimage->height, &size)) {
+                send_bgimage_to_gpu(OPT(background_image_layout), global_state.bgimage);
+            }
+        }
+        if (global_state.bgimage->texture_id) {
+            ans->bgimage = global_state.bgimage;
+            ans->bgimage->refcnt++;
+        }
+    }
+
     ans->font_sz_in_pts = global_state.font_sz_in_pts;
     END_WITH_OS_WINDOW_REFS
     return ans;
@@ -92,7 +140,7 @@ add_tab(id_type os_window_id) {
     WITH_OS_WINDOW(os_window_id)
         make_os_window_context_current(os_window);
         ensure_space_for(os_window, tabs, Tab, os_window->num_tabs + 1, capacity, 1, true);
-        memset(os_window->tabs + os_window->num_tabs, 0, sizeof(Tab));
+        zero_at_i(os_window->tabs, os_window->num_tabs);
         os_window->tabs[os_window->num_tabs].id = ++global_state.tab_id_counter;
         os_window->tabs[os_window->num_tabs].border_rects.vao_idx = create_border_vao();
         return os_window->tabs[os_window->num_tabs++].id;
@@ -100,18 +148,40 @@ add_tab(id_type os_window_id) {
     return 0;
 }
 
+static inline void
+create_gpu_resources_for_window(Window *w) {
+    w->render_data.vao_idx = create_cell_vao();
+    w->render_data.gvao_idx = create_graphics_vao();
+}
+
+static inline void
+release_gpu_resources_for_window(Window *w) {
+    if (w->render_data.vao_idx > -1) remove_vao(w->render_data.vao_idx);
+    w->render_data.vao_idx = -1;
+    if (w->render_data.gvao_idx > -1) remove_vao(w->render_data.gvao_idx);
+    w->render_data.gvao_idx = -1;
+}
+
+static inline void
+initialize_window(Window *w, PyObject *title, bool init_gpu_resources) {
+    w->id = ++global_state.window_id_counter;
+    w->visible = true;
+    w->title = title;
+    Py_XINCREF(title);
+    if (init_gpu_resources) create_gpu_resources_for_window(w);
+    else {
+        w->render_data.vao_idx = -1;
+        w->render_data.gvao_idx = -1;
+    }
+}
+
 static inline id_type
 add_window(id_type os_window_id, id_type tab_id, PyObject *title) {
     WITH_TAB(os_window_id, tab_id);
         ensure_space_for(tab, windows, Window, tab->num_windows + 1, capacity, 1, true);
         make_os_window_context_current(osw);
-        memset(tab->windows + tab->num_windows, 0, sizeof(Window));
-        tab->windows[tab->num_windows].id = ++global_state.window_id_counter;
-        tab->windows[tab->num_windows].visible = true;
-        tab->windows[tab->num_windows].title = title;
-        tab->windows[tab->num_windows].render_data.vao_idx = create_cell_vao();
-        tab->windows[tab->num_windows].render_data.gvao_idx = create_graphics_vao();
-        Py_INCREF(tab->windows[tab->num_windows].title);
+        zero_at_i(tab->windows, tab->num_windows);
+        initialize_window(tab->windows + tab->num_windows, title, true);
         return tab->windows[tab->num_windows++].id;
     END_WITH_TAB;
     return 0;
@@ -134,12 +204,12 @@ update_window_title(id_type os_window_id, id_type tab_id, id_type window_id, PyO
 static inline void
 destroy_window(Window *w) {
     Py_CLEAR(w->render_data.screen); Py_CLEAR(w->title);
-    remove_vao(w->render_data.vao_idx); remove_vao(w->render_data.gvao_idx);
+    release_gpu_resources_for_window(w);
 }
 
 static inline void
 remove_window_inner(Tab *tab, id_type id) {
-    REMOVER(tab->windows, id, tab->num_windows, Window, destroy_window, tab->capacity);
+    REMOVER(tab->windows, id, tab->num_windows, destroy_window, tab->capacity);
 }
 
 static inline void
@@ -150,6 +220,71 @@ remove_window(id_type os_window_id, id_type tab_id, id_type id) {
     END_WITH_TAB;
 }
 
+typedef struct {
+    unsigned int num_windows, capacity;
+    Window *windows;
+} DetachedWindows;
+
+static DetachedWindows detached_windows = {0};
+
+
+static void
+add_detached_window(Window *w) {
+    ensure_space_for(&detached_windows, windows, Window, detached_windows.num_windows + 1, capacity, 8, true);
+    memcpy(detached_windows.windows + detached_windows.num_windows++, w, sizeof(Window));
+}
+
+static inline void
+detach_window(id_type os_window_id, id_type tab_id, id_type id) {
+    WITH_TAB(os_window_id, tab_id);
+        for (size_t i = 0; i < tab->num_windows; i++) {
+            if (tab->windows[i].id == id) {
+                make_os_window_context_current(osw);
+                release_gpu_resources_for_window(&tab->windows[i]);
+                add_detached_window(tab->windows + i);
+                zero_at_i(tab->windows, i);
+                remove_i_from_array(tab->windows, i, tab->num_windows);
+                break;
+            }
+        }
+    END_WITH_TAB;
+}
+
+
+static inline void
+resize_screen(OSWindow *os_window, Screen *screen, bool has_graphics) {
+    if (screen) {
+        screen->cell_size.width = os_window->fonts_data->cell_width;
+        screen->cell_size.height = os_window->fonts_data->cell_height;
+        screen_dirty_sprite_positions(screen);
+        if (has_graphics) screen_rescale_images(screen);
+    }
+}
+
+static inline void
+attach_window(id_type os_window_id, id_type tab_id, id_type id) {
+    WITH_TAB(os_window_id, tab_id);
+        for (size_t i = 0; i < detached_windows.num_windows; i++) {
+            if (detached_windows.windows[i].id == id) {
+                ensure_space_for(tab, windows, Window, tab->num_windows + 1, capacity, 1, true);
+                Window *w = tab->windows + tab->num_windows++;
+                memcpy(w, detached_windows.windows + i, sizeof(Window));
+                zero_at_i(detached_windows.windows, i);
+                remove_i_from_array(detached_windows.windows, i, detached_windows.num_windows);
+                make_os_window_context_current(osw);
+                create_gpu_resources_for_window(w);
+                if (
+                    w->render_data.screen->cell_size.width != osw->fonts_data->cell_width ||
+                    w->render_data.screen->cell_size.height != osw->fonts_data->cell_height
+                ) resize_screen(osw, w->render_data.screen, true);
+                else screen_dirty_sprite_positions(w->render_data.screen);
+                w->render_data.screen->reload_all_gpu_data = true;
+                break;
+            }
+        }
+    END_WITH_TAB;
+}
+
 static inline void
 destroy_tab(Tab *tab) {
     for (size_t i = tab->num_windows; i > 0; i--) remove_window_inner(tab, tab->windows[i - 1].id);
@@ -161,7 +296,7 @@ destroy_tab(Tab *tab) {
 static inline void
 remove_tab_inner(OSWindow *os_window, id_type id) {
     make_os_window_context_current(os_window);
-    REMOVER(os_window->tabs, id, os_window->num_tabs, Tab, destroy_tab, os_window->capacity);
+    REMOVER(os_window->tabs, id, os_window->num_tabs, destroy_tab, os_window->capacity);
 }
 
 static inline void
@@ -179,8 +314,12 @@ destroy_os_window_item(OSWindow *w) {
     }
     Py_CLEAR(w->window_title); Py_CLEAR(w->tab_bar_render_data.screen);
     if (w->offscreen_texture_id) free_texture(&w->offscreen_texture_id);
+    if (w->offscreen_framebuffer) free_framebuffer(&w->offscreen_framebuffer);
     remove_vao(w->tab_bar_render_data.vao_idx);
+    remove_vao(w->gvao_idx);
     free(w->tabs); w->tabs = NULL;
+    free_bgimage(&w->bgimage, true);
+    w->bgimage = NULL;
 }
 
 bool
@@ -192,7 +331,7 @@ remove_os_window(id_type os_window_id) {
     END_WITH_OS_WINDOW
     if (found) {
         WITH_OS_WINDOW_REFS
-            REMOVER(global_state.os_windows, os_window_id, global_state.num_os_windows, OSWindow, destroy_os_window_item, global_state.capacity);
+            REMOVER(global_state.os_windows, os_window_id, global_state.num_os_windows, destroy_os_window_item, global_state.capacity);
         END_WITH_OS_WINDOW_REFS
         update_os_window_references();
     }
@@ -249,7 +388,7 @@ add_borders_rect(id_type os_window_id, id_type tab_id, uint32_t left, uint32_t t
 
 void
 os_window_regions(OSWindow *os_window, Region *central, Region *tab_bar) {
-    if (os_window->num_tabs > 1) {
+    if (!global_state.tab_bar_hidden && os_window->num_tabs >= OPT(tab_bar_min_tabs)) {
         switch(OPT(tab_bar_edge)) {
             case TOP_EDGE:
                 central->left = 0; central->top = os_window->fonts_data->cell_height; central->right = os_window->viewport_width - 1;
@@ -265,7 +404,7 @@ os_window_regions(OSWindow *os_window, Region *central, Region *tab_bar) {
                 break;
         }
     } else {
-        memset(tab_bar, 0, sizeof(Region));
+        zero_at_ptr(tab_bar);
         central->left = 0; central->top = 0; central->right = os_window->viewport_width - 1;
         central->bottom = os_window->viewport_height - 1;
     }
@@ -289,22 +428,26 @@ os_window_regions(OSWindow *os_window, Region *central, Region *tab_bar) {
 #define KK5I(name) PYWRAP1(name) { id_type a, b; unsigned int c, d, e, f, g; PA("KKIIIII", &a, &b, &c, &d, &e, &f, &g); name(a, b, c, d, e, f, g); Py_RETURN_NONE; }
 #define BOOL_SET(name) PYWRAP1(set_##name) { global_state.name = PyObject_IsTrue(args); Py_RETURN_NONE; }
 
+static color_type default_color = 0;
+
 static inline color_type
 color_as_int(PyObject *color) {
+    if (color == Py_None && default_color) return default_color;
     if (!PyTuple_Check(color)) { PyErr_SetString(PyExc_TypeError, "Not a color tuple"); return 0; }
 #define I(n, s) ((PyLong_AsUnsignedLong(PyTuple_GET_ITEM(color, n)) & 0xff) << s)
     return (I(0, 16) | I(1, 8) | I(2, 0)) & 0xffffff;
 #undef I
 }
 
-static inline double
-repaint_delay(PyObject *val) {
-    return (double)(PyLong_AsUnsignedLong(val)) / 1000.0;
+static inline monotonic_t
+parse_s_double_to_monotonic_t(PyObject *val) {
+    return s_double_to_monotonic_t(PyFloat_AsDouble(val));
 }
 
-#define dict_iter(d) { \
-    PyObject *key, *value; Py_ssize_t pos = 0; \
-    while (PyDict_Next(d, &pos, &key, &value))
+static inline monotonic_t
+parse_ms_long_to_monotonic_t(PyObject *val) {
+    return ms_to_monotonic_t(PyLong_AsUnsignedLong(val));
+}
 
 static int kitty_mod = 0;
 
@@ -321,6 +464,87 @@ convert_mods(PyObject *obj) {
     return resolve_mods(PyLong_AsLong(obj));
 }
 
+static WindowTitleIn
+window_title_in(PyObject *title_in) {
+    const char *in = PyUnicode_AsUTF8(title_in);
+    switch(in[0]) {
+        case 'a': return ALL;
+        case 'w': return WINDOW;
+        case 'm': return MENUBAR;
+        case 'n': return NONE;
+        default: break;
+    }
+    return ALL;
+}
+
+static BackgroundImageLayout
+bglayout(PyObject *layout_name) {
+    const char *name = PyUnicode_AsUTF8(layout_name);
+    switch(name[0]) {
+        case 't': return TILING;
+        case 'm': return MIRRORED;
+        case 's': return SCALED;
+        default: break;
+    }
+    return TILING;
+}
+
+static void
+background_image(PyObject *src) {
+    if (OPT(background_image)) free(OPT(background_image));
+    OPT(background_image) = NULL;
+    if (src == Py_None || !PyUnicode_Check(src)) return;
+    Py_ssize_t sz;
+    const char *s = PyUnicode_AsUTF8AndSize(src, &sz);
+    OPT(background_image) = calloc(sz + 1, 1);
+    if (OPT(background_image)) memcpy(OPT(background_image), s, sz);
+}
+
+
+static MouseShape
+pointer_shape(PyObject *shape_name) {
+    const char *name = PyUnicode_AsUTF8(shape_name);
+    switch(name[0]) {
+        case 'a': return ARROW;
+        case 'h': return HAND;
+        case 'b': return BEAM;
+        default: break;
+    }
+    return BEAM;
+}
+
+static inline void
+free_url_prefixes(void) {
+    OPT(url_prefixes).num = 0;
+    OPT(url_prefixes).max_prefix_len = 0;
+    if (OPT(url_prefixes).values) {
+        free(OPT(url_prefixes.values));
+        OPT(url_prefixes).values = NULL;
+    }
+}
+
+static void
+set_url_prefixes(PyObject *up) {
+    free_url_prefixes();
+    OPT(url_prefixes).values = calloc(PyTuple_GET_SIZE(up), sizeof(UrlPrefix));
+    if (!OPT(url_prefixes).values) { PyErr_NoMemory(); return; }
+    OPT(url_prefixes).num = PyTuple_GET_SIZE(up);
+    for (size_t i = 0; i < OPT(url_prefixes).num; i++) {
+        PyObject *t = PyTuple_GET_ITEM(up, i);
+        if (!PyUnicode_Check(t)) { PyErr_SetString(PyExc_TypeError, "url_prefixes must be strings"); return; }
+        OPT(url_prefixes).values[i].len = MIN(arraysz(OPT(url_prefixes).values[i].string) - 1, (size_t)PyUnicode_GET_LENGTH(t));
+        int kind = PyUnicode_KIND(t);
+        OPT(url_prefixes).max_prefix_len = MAX(OPT(url_prefixes).max_prefix_len, OPT(url_prefixes).values[i].len);
+        for (size_t x = 0; x < OPT(url_prefixes).values[i].len; x++) {
+            OPT(url_prefixes).values[i].string[x] = PyUnicode_READ(kind, PyUnicode_DATA(t), x);
+        }
+    }
+}
+
+#define dict_iter(d) { \
+    PyObject *key, *value; Py_ssize_t pos = 0; \
+    while (PyDict_Next(d, &pos, &key, &value))
+
 static inline void
 set_special_keys(PyObject *dict) {
     dict_iter(dict) {
@@ -332,6 +556,11 @@ set_special_keys(PyObject *dict) {
     }}
 }
 
+static inline float
+PyFloat_AsFloat(PyObject *o) {
+    return (float)PyFloat_AsDouble(o);
+}
+
 PYWRAP0(next_window_id) {
     return PyLong_FromUnsignedLongLong(global_state.window_id_counter + 1);
 }
@@ -351,48 +580,88 @@ PYWRAP1(set_options) {
     int is_wayland = 0, debug_gl = 0, debug_font_fallback = 0;
     PA("O|ppp", &opts, &is_wayland, &debug_gl, &debug_font_fallback);
     global_state.is_wayland = is_wayland ? true : false;
+#ifdef __APPLE__
+    global_state.has_render_frames = true;
+#endif
+    if (global_state.is_wayland) global_state.has_render_frames = true;
     global_state.debug_gl = debug_gl ? true : false;
     global_state.debug_font_fallback = debug_font_fallback ? true : false;
 #define GA(name) ret = PyObject_GetAttrString(opts, #name); if (ret == NULL) return NULL;
-#define S(name, convert) { GA(name); global_state.opts.name = convert(ret); Py_DECREF(ret); if (PyErr_Occurred()) return NULL; }
-    GA(kitty_mod);
-    kitty_mod = PyLong_AsLong(ret); Py_CLEAR(ret); if (PyErr_Occurred()) return NULL;
-    S(visual_bell_duration, PyFloat_AsDouble);
+#define SS(name, dest, convert) { GA(name); dest = convert(ret); Py_DECREF(ret); if (PyErr_Occurred()) return NULL; }
+#define S(name, convert) SS(name, OPT(name), convert)
+    SS(kitty_mod, kitty_mod, PyLong_AsLong);
+    S(hide_window_decorations, PyLong_AsUnsignedLong);
+    S(visual_bell_duration, parse_s_double_to_monotonic_t);
     S(enable_audio_bell, PyObject_IsTrue);
     S(focus_follows_mouse, PyObject_IsTrue);
-    S(cursor_blink_interval, PyFloat_AsDouble);
-    S(cursor_stop_blinking_after, PyFloat_AsDouble);
-    S(background_opacity, PyFloat_AsDouble);
-    S(dim_opacity, PyFloat_AsDouble);
+    S(cursor_blink_interval, parse_s_double_to_monotonic_t);
+    S(cursor_stop_blinking_after, parse_s_double_to_monotonic_t);
+    S(background_opacity, PyFloat_AsFloat);
+    S(background_image_layout, bglayout);
+    S(background_tint, PyFloat_AsFloat);
+    S(background_image_linear, PyObject_IsTrue);
+    S(dim_opacity, PyFloat_AsFloat);
     S(dynamic_background_opacity, PyObject_IsTrue);
-    S(inactive_text_alpha, PyFloat_AsDouble);
-    S(window_padding_width, PyFloat_AsDouble);
+    S(inactive_text_alpha, PyFloat_AsFloat);
+    S(window_padding_width, PyFloat_AsFloat);
+    S(scrollback_pager_history_size, PyLong_AsUnsignedLong);
     S(cursor_shape, PyLong_AsLong);
+    S(cursor_beam_thickness, PyFloat_AsFloat);
+    S(cursor_underline_thickness, PyFloat_AsFloat);
     S(url_style, PyLong_AsUnsignedLong);
     S(tab_bar_edge, PyLong_AsLong);
-    S(mouse_hide_wait, PyFloat_AsDouble);
+    S(mouse_hide_wait, parse_s_double_to_monotonic_t);
     S(wheel_scroll_multiplier, PyFloat_AsDouble);
+    S(touch_scroll_multiplier, PyFloat_AsDouble);
     S(open_url_modifiers, convert_mods);
     S(rectangle_select_modifiers, convert_mods);
-    S(click_interval, PyFloat_AsDouble);
+    S(terminal_select_modifiers, convert_mods);
+    S(click_interval, parse_s_double_to_monotonic_t);
+    S(resize_debounce_time, parse_s_double_to_monotonic_t);
+    S(mark1_foreground, color_as_int);
+    S(mark1_background, color_as_int);
+    S(mark2_foreground, color_as_int);
+    S(mark2_background, color_as_int);
+    S(mark3_foreground, color_as_int);
+    S(mark3_background, color_as_int);
     S(url_color, color_as_int);
     S(background, color_as_int);
+    S(foreground, color_as_int);
+    default_color = 0x00ff00;
     S(active_border_color, color_as_int);
+    default_color = 0;
     S(inactive_border_color, color_as_int);
     S(bell_border_color, color_as_int);
-    S(repaint_delay, repaint_delay);
-    S(input_delay, repaint_delay);
+    S(repaint_delay, parse_ms_long_to_monotonic_t);
+    S(input_delay, parse_ms_long_to_monotonic_t);
     S(sync_to_monitor, PyObject_IsTrue);
     S(close_on_child_death, PyObject_IsTrue);
     S(window_alert_on_bell, PyObject_IsTrue);
-    S(macos_option_as_alt, PyObject_IsTrue);
+    S(macos_option_as_alt, PyLong_AsUnsignedLong);
     S(macos_traditional_fullscreen, PyObject_IsTrue);
-    S(macos_hide_titlebar, PyObject_IsTrue);
     S(macos_quit_when_last_window_closed, PyObject_IsTrue);
+    S(macos_show_window_title_in, window_title_in);
     S(macos_window_resizable, PyObject_IsTrue);
-    S(x11_hide_window_decorations, PyObject_IsTrue);
     S(macos_hide_from_tasks, PyObject_IsTrue);
-    S(macos_thicken_font, PyFloat_AsDouble);
+    S(macos_thicken_font, PyFloat_AsFloat);
+    S(tab_bar_min_tabs, PyLong_AsUnsignedLong);
+    S(disable_ligatures, PyLong_AsLong);
+    S(force_ltr, PyObject_IsTrue);
+    S(resize_draw_strategy, PyLong_AsLong);
+    S(resize_in_steps, PyObject_IsTrue);
+    S(pointer_shape_when_grabbed, pointer_shape);
+
+    GA(tab_bar_style);
+    global_state.tab_bar_hidden = PyUnicode_CompareWithASCIIString(ret, "hidden") == 0 ? true: false;
+    Py_CLEAR(ret);
+    if (PyErr_Occurred()) return NULL;
+
+    PyObject *up = PyObject_GetAttrString(opts, "url_prefixes");
+    if (up == NULL) return NULL;
+    if (!PyTuple_Check(up)) { PyErr_SetString(PyExc_TypeError, "url_prefixes must be a tuple"); return NULL; }
+    set_url_prefixes(up);
+    Py_DECREF(up);
+    if (PyErr_Occurred()) return NULL;
 
     PyObject *chars = PyObject_GetAttrString(opts, "select_by_word_characters");
     if (chars == NULL) return NULL;
@@ -407,6 +676,8 @@ PYWRAP1(set_options) {
     GA(sequence_map); set_special_keys(ret);
     Py_DECREF(ret); if (PyErr_Occurred()) return NULL;
 
+    GA(background_image); background_image(ret); Py_CLEAR(ret);
+
 #define read_adjust(name) { \
     PyObject *al = PyObject_GetAttrString(opts, #name); \
     if (PyFloat_Check(al)) { \
@@ -422,6 +693,7 @@ PYWRAP1(set_options) {
     read_adjust(adjust_column_width);
 #undef read_adjust
 #undef S
+#undef SS
     Py_RETURN_NONE;
 }
 
@@ -489,6 +761,15 @@ end:
 }
 
 
+PYWRAP1(os_window_has_background_image) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        if (os_window->bgimage && os_window->bgimage->texture_id > 0) { Py_RETURN_TRUE; }
+    END_WITH_OS_WINDOW
+    Py_RETURN_FALSE;
+}
+
 PYWRAP1(mark_os_window_for_close) {
     id_type os_window_id;
     int yes = 1;
@@ -626,16 +907,6 @@ PYWRAP1(global_font_size) {
     return Py_BuildValue("d", global_state.font_sz_in_pts);
 }
 
-static inline void
-resize_screen(OSWindow *os_window, Screen *screen, bool has_graphics) {
-    if (screen) {
-        screen->cell_size.width = os_window->fonts_data->cell_width;
-        screen->cell_size.height = os_window->fonts_data->cell_height;
-        screen_dirty_sprite_positions(screen);
-        if (has_graphics) screen_rescale_images(screen);
-    }
-}
-
 PYWRAP1(os_window_font_size) {
     id_type os_window_id;
     int force = 0;
@@ -655,6 +926,7 @@ PYWRAP1(os_window_font_size) {
                     resize_screen(os_window, w->render_data.screen, true);
                 }
             }
+            if (OPT(resize_in_steps)) os_window_update_size_increments(os_window);
         }
         return Py_BuildValue("d", os_window->font_sz_in_pts);
     END_WITH_OS_WINDOW
@@ -668,6 +940,14 @@ PYWRAP1(set_boss) {
     Py_RETURN_NONE;
 }
 
+PYWRAP0(get_boss) {
+    if (global_state.boss) {
+        Py_INCREF(global_state.boss);
+        return global_state.boss;
+    }
+    Py_RETURN_NONE;
+}
+
 PYWRAP1(patch_global_colors) {
     PyObject *spec;
     int configured;
@@ -675,25 +955,95 @@ PYWRAP1(patch_global_colors) {
 #define P(name) { \
     PyObject *val = PyDict_GetItemString(spec, #name); \
     if (val) { \
-		global_state.opts.name = PyLong_AsLong(val); \
-	} \
+        OPT(name) = PyLong_AsLong(val); \
+    } \
 }
     P(active_border_color); P(inactive_border_color); P(bell_border_color);
     if (configured) {
         P(background); P(url_color);
+        P(mark1_background); P(mark1_foreground); P(mark2_background); P(mark2_foreground);
+        P(mark3_background); P(mark3_foreground);
     }
     if (PyErr_Occurred()) return NULL;
     Py_RETURN_NONE;
 }
 
+static PyObject*
+pyset_background_image(PyObject *self UNUSED, PyObject *args) {
+    const char *path;
+    PyObject *layout_name = NULL;
+    PyObject *os_window_ids;
+    int configured = 0;
+    PA("zO!|pU", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name);
+    size_t size;
+    BackgroundImageLayout layout = layout_name ? bglayout(layout_name) : OPT(background_image_layout);
+    BackgroundImage *bgimage = NULL;
+    if (path) {
+        bgimage = calloc(1, sizeof(BackgroundImage));
+        if (!bgimage) return PyErr_NoMemory();
+        if (!png_path_to_bitmap(path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size)) {
+            PyErr_Format(PyExc_ValueError, "Failed to load image from: %s", path);
+            free(bgimage);
+            return NULL;
+        }
+        send_bgimage_to_gpu(layout, bgimage);
+        bgimage->refcnt++;
+    }
+    if (configured) {
+        free_bgimage(&global_state.bgimage, true);
+        global_state.bgimage = bgimage;
+        if (bgimage) bgimage->refcnt++;
+        OPT(background_image_layout) = layout;
+    }
+    for (Py_ssize_t i = 0; i < PyTuple_GET_SIZE(os_window_ids); i++) {
+        id_type os_window_id = PyLong_AsUnsignedLongLong(PyTuple_GET_ITEM(os_window_ids, i));
+        WITH_OS_WINDOW(os_window_id)
+            make_os_window_context_current(os_window);
+            free_bgimage(&os_window->bgimage, true);
+            os_window->bgimage = bgimage;
+            os_window->render_calls = 0;
+            if (bgimage) bgimage->refcnt++;
+        END_WITH_OS_WINDOW
+    }
+    if (bgimage) free_bgimage(&bgimage, true);
+    Py_RETURN_NONE;
+}
+
 PYWRAP0(destroy_global_data) {
     Py_CLEAR(global_state.boss);
     free(global_state.os_windows); global_state.os_windows = NULL;
     Py_RETURN_NONE;
 }
 
+static void
+destroy_mock_window(PyObject *capsule) {
+    Window *w = PyCapsule_GetPointer(capsule, "Window");
+    if (w) {
+        destroy_window(w);
+        PyMem_Free(w);
+    }
+}
+
+static PyObject*
+pycreate_mock_window(PyObject *self UNUSED, PyObject *args) {
+    Screen *screen;
+    PyObject *title = NULL;
+    if (!PyArg_ParseTuple(args, "O|U", &screen, &title)) return NULL;
+    Window *w = PyMem_Calloc(sizeof(Window), 1);
+    if (!w) return NULL;
+    Py_INCREF(screen);
+    PyObject *ans = PyCapsule_New(w, "Window", destroy_mock_window);
+    if (ans != NULL) {
+        initialize_window(w, title, false);
+        w->render_data.screen = screen;
+    }
+    return ans;
+}
+
 THREE_ID_OBJ(update_window_title)
 THREE_ID(remove_window)
+THREE_ID(detach_window)
+THREE_ID(attach_window)
 PYWRAP1(resolve_key_mods) { int mods; PA("ii", &kitty_mod, &mods); return PyLong_FromLong(resolve_mods(mods)); }
 PYWRAP1(add_tab) { return PyLong_FromUnsignedLongLong(add_tab(PyLong_AsUnsignedLongLong(args))); }
 PYWRAP1(add_window) { PyObject *title; id_type a, b; PA("KKO", &a, &b, &title); return PyLong_FromUnsignedLongLong(add_window(a, b, title)); }
@@ -721,6 +1071,8 @@ static PyMethodDef module_methods[] = {
     MW(update_window_title, METH_VARARGS),
     MW(remove_tab, METH_VARARGS),
     MW(remove_window, METH_VARARGS),
+    MW(detach_window, METH_VARARGS),
+    MW(attach_window, METH_VARARGS),
     MW(set_active_tab, METH_VARARGS),
     MW(set_active_window, METH_VARARGS),
     MW(swap_tabs, METH_VARARGS),
@@ -730,6 +1082,7 @@ static PyMethodDef module_methods[] = {
     MW(set_window_render_data, METH_VARARGS),
     MW(viewport_for_window, METH_VARARGS),
     MW(cell_size_for_window, METH_VARARGS),
+    MW(os_window_has_background_image, METH_VARARGS),
     MW(mark_os_window_for_close, METH_VARARGS),
     MW(set_titlebar_color, METH_VARARGS),
     MW(focus_os_window, METH_VARARGS),
@@ -738,14 +1091,34 @@ static PyMethodDef module_methods[] = {
     MW(background_opacity_of, METH_O),
     MW(update_window_visibility, METH_VARARGS),
     MW(global_font_size, METH_VARARGS),
+    MW(set_background_image, METH_VARARGS),
     MW(os_window_font_size, METH_VARARGS),
     MW(set_boss, METH_O),
+    MW(get_boss, METH_NOARGS),
     MW(patch_global_colors, METH_VARARGS),
+    MW(create_mock_window, METH_VARARGS),
     MW(destroy_global_data, METH_NOARGS),
 
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
+static void
+finalize(void) {
+    while(detached_windows.num_windows--) {
+        destroy_window(&detached_windows.windows[detached_windows.num_windows]);
+    }
+    if (detached_windows.windows) free(detached_windows.windows);
+    detached_windows.capacity = 0;
+    if (OPT(background_image)) free(OPT(background_image));
+    // we leak the texture here since it is not guaranteed
+    // that freeing the texture will work during shutdown and
+    // the GPU driver should take care of it when the OpenGL context is
+    // destroyed.
+    free_bgimage(&global_state.bgimage, false);
+    global_state.bgimage = NULL;
+    free_url_prefixes();
+}
+
 bool
 init_state(PyObject *module) {
     global_state.font_sz_in_pts = 11.0;
@@ -759,6 +1132,10 @@ init_state(PyObject *module) {
     if (PyStructSequence_InitType2(&RegionType, &region_desc) != 0) return false;
     Py_INCREF((PyObject *) &RegionType);
     PyModule_AddObject(module, "Region", (PyObject *) &RegionType);
+    if (Py_AtExit(finalize) != 0) {
+        PyErr_SetString(PyExc_RuntimeError, "Failed to register the state at exit handler");
+        return false;
+    }
     return true;
 }
 // }}}
