@@ -1,9 +1,10 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 from . import BaseTest
 from kitty.fast_data_types import DECAWM, IRM, Cursor, DECCOLM, DECOM
+from kitty.marks import marker_from_regex, marker_from_function
 
 
 class TestScreen(BaseTest):
@@ -293,11 +294,7 @@ class TestScreen(BaseTest):
             s.tab()
             s.draw('*')
         s.cursor_position(2, 2)
-        for col in range(2, s.columns - 1, 6):
-            for i in range(5):
-                s.draw(' ')
-            s.draw('*')
-        self.ae(str(s.line(0)), str(s.line(1)))
+        self.ae(str(s.line(0)), '\t*'*13)
 
     def test_margins(self):
         # Taken from vttest/main.c
@@ -338,7 +335,8 @@ class TestScreen(BaseTest):
                     s.cursor_position(region, s.columns), s.draw(ch.lower())
             for l in range(2, region + 2):
                 c = chr(ord('I') + l - 2)
-                self.ae(c + ' ' * (s.columns - 2) + c.lower(), str(s.line(l)))
+                before = '\t' if l % 4 == 0 else ' '
+                self.ae(c + ' ' * (s.columns - 3) + before + c.lower(), str(s.line(l)))
             s.reset_mode(DECOM)
         # Test that moving cursor outside the margins works as expected
         s = self.create_screen(10, 10)
@@ -422,13 +420,27 @@ class TestScreen(BaseTest):
             if i != 0:
                 s.carriage_return(), s.linefeed()
             s.draw(str(i) * s.columns)
-        s.start_selection(0, 0, False)
-        s.update_selection(4, 4, True)
+        s.start_selection(0, 0)
+        s.update_selection(4, 4)
         expected = ('55555', '\n66666', '\n77777', '\n88888', '\n99999')
         self.ae(s.text_for_selection(), expected)
         s.scroll(2, True)
         self.ae(s.text_for_selection(), expected)
 
+    def test_variation_selectors(self):
+        s = self.create_screen()
+        s.draw('\U0001f610')
+        self.ae(s.cursor.x, 2)
+        s.carriage_return(), s.linefeed()
+        s.draw('\U0001f610\ufe0e')
+        self.ae(s.cursor.x, 1)
+        s.carriage_return(), s.linefeed()
+        s.draw('\u25b6')
+        self.ae(s.cursor.x, 1)
+        s.carriage_return(), s.linefeed()
+        s.draw('\u25b6\ufe0f')
+        self.ae(s.cursor.x, 2)
+
     def test_serialize(self):
         s = self.create_screen()
         s.draw('ab' * s.columns)
@@ -441,4 +453,38 @@ class TestScreen(BaseTest):
             return ''.join(d)
 
         self.ae(as_text(), 'ababababab\nc\n\n')
-        self.ae(as_text(True), '\x1b[0mababa\x1b[0mbabab\n\x1b[0mc\n\n')
+        self.ae(as_text(True), '\x1b[mababa\x1b[mbabab\n\x1b[mc\n\n')
+
+    def test_user_marking(self):
+        s = self.create_screen()
+        s.draw('abaa')
+        s.carriage_return(), s.linefeed()
+        s.draw('xyxyx')
+        s.set_marker(marker_from_regex('a', 3))
+        self.ae(s.marked_cells(), [(0, 0, 3), (2, 0, 3), (3, 0, 3)])
+        s.set_marker()
+        self.ae(s.marked_cells(), [])
+
+        def mark_x(text):
+            col = 0
+            for i, c in enumerate(text):
+                if c == 'x':
+                    col += 1
+                    yield i, i, col
+
+        s.set_marker(marker_from_function(mark_x))
+        self.ae(s.marked_cells(), [(0, 1, 1), (2, 1, 2), (4, 1, 3)])
+        s = self.create_screen(lines=5, scrollback=10)
+        for i in range(15):
+            s.draw(str(i))
+            if i != 14:
+                s.carriage_return(), s.linefeed()
+        s.set_marker(marker_from_regex(r'\d+', 3))
+        for i in range(10):
+            self.assertTrue(s.scroll_to_next_mark())
+            self.ae(s.scrolled_by, i + 1)
+        self.ae(s.scrolled_by, 10)
+        for i in range(10):
+            self.assertTrue(s.scroll_to_next_mark(0, False))
+            self.ae(s.scrolled_by, 10 - i - 1)
+        self.ae(s.scrolled_by, 0)
